@@ -1,6 +1,6 @@
 /*****************************************************************************
 
-  Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+  Copyright (c) 2001, 2002 Zope Foundation and Contributors.
   All Rights Reserved.
 
   This software is subject to the provisions of the Zope Public License,
@@ -12,7 +12,7 @@
 
  ****************************************************************************/
 
-#define BUCKETTEMPLATE_C "$Id: BucketTemplate.c 103090 2009-08-22 17:46:16Z jim $\n"
+#define BUCKETTEMPLATE_C "$Id: BucketTemplate.c 113734 2010-06-21 15:33:46Z ctheune $\n"
 
 /* Use BUCKET_SEARCH to find the index at which a key belongs.
  * INDEX    An int lvalue to hold the index i such that KEY belongs at
@@ -1582,6 +1582,10 @@ static struct PyMethodDef Bucket_methods[] = {
     {"iteritems", (PyCFunction) Bucket_iteritems,    METH_KEYWORDS,
      "B.iteritems([min[,max]]) -> an iterator over the (key, value) items of B"},
 
+#ifdef EXTRA_BUCKET_METHODS
+    EXTRA_BUCKET_METHODS
+#endif
+
 #ifdef PERSISTENT
     {"_p_resolveConflict", (PyCFunction) bucket__p_resolveConflict,
      METH_VARARGS,
