@@ -11,12 +11,8 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-"""XXX short summary goes here.
-
-$Id: tests.py 113733 2010-06-21 15:32:58Z ctheune $
-"""
-import re, unittest
-from zope.testing import doctest, renormalizing
+import doctest, re, unittest
+from zope.testing import renormalizing
 
 def test_suite():
     return unittest.TestSuite((
