@@ -13,7 +13,7 @@
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py 117033 2010-09-28 20:27:49Z jim $"""
+$Id: Connection.py 121409 2011-04-12 13:50:07Z jim $"""
 
 import logging
 import sys
@@ -21,7 +21,6 @@ import tempfile
 import threading
 import warnings
 import os
-import shutil
 import time
 
 from persistent import PickleCache
@@ -48,7 +47,7 @@ from ZODB.POSException import InvalidObjectReference, ConnectionStateError
 from ZODB.POSException import ConflictError, ReadConflictError
 from ZODB.POSException import Unsupported, ReadOnlyHistoryError
 from ZODB.POSException import POSKeyError
-from ZODB.serialize import ObjectWriter, ObjectReader, myhasattr
+from ZODB.serialize import ObjectWriter, ObjectReader
 from ZODB.utils import p64, u64, z64, oid_repr, positive_id
 from ZODB import utils
 
@@ -88,6 +87,7 @@ class Connection(ExportImport, object):
         self._debug_info = ()
 
         self._db = db
+        self.large_record_size = db.large_record_size
 
         # historical connection
         self.before = before
@@ -104,7 +104,7 @@ class Connection(ExportImport, object):
             self._mvcc_storage = False
 
         self._normal_storage = self._storage = storage
-        self.new_oid = storage.new_oid
+        self.new_oid = db.new_oid
         self._savepoint_storage = None
 
         # Do we need to join a txn manager?
@@ -133,6 +133,10 @@ class Connection(ExportImport, object):
         # of this list are either in _cache or in _added.
         self._registered_objects = []
 
+        # ids and serials of objects for which readCurrent was called
+        # in a transaction.
+        self._readCurrent = {}
+
         # Dict of oid->obj added explicitly through add(). Used as a
         # preliminary cache until commit time when objects are all moved
         # to the real _cache. The objects are moved to _creating at
@@ -214,7 +218,7 @@ class Connection(ExportImport, object):
                             " added to a Connection.", obj)
         elif obj._p_jar is None:
             assert obj._p_oid is None
-            oid = obj._p_oid = self._storage.new_oid()
+            oid = obj._p_oid = self.new_oid()
             obj._p_jar = self
             if self._added_during_commit is not None:
                 self._added_during_commit.append(obj)
@@ -241,24 +245,16 @@ class Connection(ExportImport, object):
         if obj is not None:
             return obj
 
-        # This appears to be an MVCC violation because we are loading
-        # the must recent data when perhaps we shouldnt. The key is
-        # that we are only creating a ghost!
-        # A disadvantage to this optimization is that _p_serial cannot be
-        # trusted until the object has been loaded, which affects both MVCC
-        # and historical connections.
         p, serial = self._storage.load(oid, '')
         obj = self._reader.getGhost(p)
 
         # Avoid infiniate loop if obj tries to load its state before
         # it is added to the cache and it's state refers to it.
+        # (This will typically be the case for non-ghostifyable objects,
+        # like persistent caches.)
         self._pre_cache[oid] = obj
-        obj._p_oid = oid
-        obj._p_jar = self
-        obj._p_changed = None
-        obj._p_serial = serial
+        self._cache.new_ghost(oid, obj)
         self._pre_cache.pop(oid)
-        self._cache[oid] = obj
         return obj
 
     def cacheMinimize(self):
@@ -325,6 +321,10 @@ class Connection(ExportImport, object):
                 # get back here.
         else:
             self.opened = None
+            
+        am = self._db._activity_monitor
+        if am is not None:
+            am.closedConnection(self)        
 
     def db(self):
         """Returns a handle to the database this connection belongs to."""
@@ -438,6 +438,7 @@ class Connection(ExportImport, object):
         if self._savepoint_storage is not None:
             self._abort_savepoint()
 
+        self._invalidate_creating()
         self._tpc_cleanup()
 
     def _abort(self):
@@ -565,6 +566,14 @@ class Connection(ExportImport, object):
         else:
             self._commit(transaction)
 
+        for oid, serial in self._readCurrent.iteritems():
+            try:
+                self._storage.checkCurrentSerialInTransaction(
+                    oid, serial, transaction)
+            except ConflictError:
+                self._cache.invalidate(oid)
+                raise
+
     def _commit(self, transaction):
         """Commit changes to an object"""
 
@@ -645,7 +654,10 @@ class Connection(ExportImport, object):
                     and not hasattr(obj, '_p_resolveConflict')):
                     raise ConflictError(object=obj)
                 self._modified.append(oid)
+
             p = writer.serialize(obj)  # This calls __getstate__ of obj
+            if len(p) >= self.large_record_size:
+                warnings.warn(large_object_message % (obj.__class__, len(p)))
 
             if isinstance(obj, Blob):
                 if not IBlobStorage.providedBy(self._storage):
@@ -654,8 +666,12 @@ class Connection(ExportImport, object):
                         repr(self._storage))
                 if obj.opened():
                     raise ValueError("Can't commit with opened blobs.")
-                s = self._storage.storeBlob(oid, serial, p,
-                                            obj._uncommitted(),
+                blobfilename = obj._uncommitted()
+                if blobfilename is None:
+                    assert serial is not None # See _uncommitted
+                    self._modified.pop() # not modified
+                    continue
+                s = self._storage.storeBlob(oid, serial, p, blobfilename,
                                             '', transaction)
                 # we invalidate the object here in order to ensure
                 # that that the next attribute access of its name
@@ -682,32 +698,16 @@ class Connection(ExportImport, object):
             self._cache.update_object_size_estimation(oid, len(p))
             obj._p_estimated_size = len(p)
 
-            self._handle_serial(s, oid)
+            self._handle_serial(oid, s)
 
-    def _handle_serial(self, store_return, oid=None, change=1):
-        """Handle the returns from store() and tpc_vote() calls."""
+    def _handle_serial(self, oid, serial, change=True):
 
-        # These calls can return different types depending on whether
-        # ZEO is used.  ZEO uses asynchronous returns that may be
-        # returned in batches by the ClientStorage.  ZEO1 can also
-        # return an exception object and expect that the Connection
-        # will raise the exception.
+        # if we write an object, we don't want to check if it was read
+        # while current.  This is a convenient choke point to do this.
+        self._readCurrent.pop(oid, None)
 
-        # When conflict resolution occurs, the object state held by
-        # the connection does not match what is written to the
-        # database.  Invalidate the object here to guarantee that
-        # the new state is read the next time the object is used.
-
-        if not store_return:
+        if not serial:
             return
-        if isinstance(store_return, str):
-            assert oid is not None
-            self._handle_one_serial(oid, store_return, change)
-        else:
-            for oid, serial in store_return:
-                self._handle_one_serial(oid, serial, change)
-
-    def _handle_one_serial(self, oid, serial, change):
         if not isinstance(serial, str):
             raise serial
         obj = self._cache.get(oid, None)
@@ -759,6 +759,7 @@ class Connection(ExportImport, object):
         """Disown any objects newly saved in an uncommitted transaction."""
         if creating is None:
             creating = self._creating
+            self._creating = {}
 
         for oid in creating:
             o = self._cache.get(oid)
@@ -769,7 +770,6 @@ class Connection(ExportImport, object):
                 del o._p_jar
                 del o._p_oid
 
-        creating.clear()
 
     def tpc_vote(self, transaction):
         """Verify that a data manager can commit the transaction."""
@@ -777,8 +777,16 @@ class Connection(ExportImport, object):
             vote = self._storage.tpc_vote
         except AttributeError:
             return
-        s = vote(transaction)
-        self._handle_serial(s)
+        try:
+            s = vote(transaction)
+        except ReadConflictError, v:
+            if v.oid:
+                self._cache.invalidate(v.oid)
+            raise
+
+        if s:
+            for oid, serial in s:
+                self._handle_serial(oid, serial)
 
     def tpc_finish(self, transaction):
         """Indicate confirmation that the transaction is done."""
@@ -816,6 +824,7 @@ class Connection(ExportImport, object):
     # pending invalidations regardless.  Of course this should only be
     # called at transaction boundaries.
     def _storage_sync(self, *ignored):
+        self._readCurrent.clear()
         sync = getattr(self._storage, 'sync', 0)
         if sync:
             sync()
@@ -865,16 +874,9 @@ class Connection(ExportImport, object):
         # load.  We can only be sure about invalidations after the
         # load.
 
-        # If an object has been invalidated, there are several cases
-        # to consider:
-        # 1. Check _p_independent()
-        # 2. Try MVCC
-        # 3. Raise ConflictError.
-
-        # Does anything actually use _p_independent()?  It would simplify
-        # the code if we could drop support for it.
-        # (BTrees.Length does.)
-
+        # If an object has been invalidated, among the cases to consider:
+        # - Try MVCC
+        # - Raise ConflictError.
 
         if self.before is not None:
             # Load data that was current before the time we have.
@@ -892,9 +894,7 @@ class Connection(ExportImport, object):
             if self._invalidatedCache:
                 raise ReadConflictError()
 
-            if (obj._p_oid in self._invalidated and
-                    not myhasattr(obj, "_p_independent")):
-                # If the object has _p_independent(), we will handle it below.
+            if (obj._p_oid in self._invalidated):
                 self._load_before_or_conflict(obj)
                 return
 
@@ -908,13 +908,8 @@ class Connection(ExportImport, object):
                 self._inv_lock.release()
 
             if invalid:
-                if myhasattr(obj, "_p_independent"):
-                    # This call will raise a ReadConflictError if something
-                    # goes wrong
-                    self._handle_independent(obj)
-                else:
-                    self._load_before_or_conflict(obj)
-                    return
+                self._load_before_or_conflict(obj)
+                return
 
         self._reader.setGhostState(obj, p)
         obj._p_serial = serial
@@ -963,25 +958,6 @@ class Connection(ExportImport, object):
 
         return True
 
-    def _handle_independent(self, obj):
-        # Helper method for setstate() handles possibly independent objects
-        # Call _p_independent(), if it returns True, setstate() wins.
-        # Otherwise, raise a ConflictError.
-
-        if obj._p_independent():
-            self._inv_lock.acquire()
-            try:
-                try:
-                    self._invalidated.remove(obj._p_oid)
-                except KeyError:
-                    pass
-            finally:
-                self._inv_lock.release()
-        else:
-            self._conflicts[obj._p_oid] = 1
-            self._register(obj)
-            raise ReadConflictError(object=obj)
-
     def register(self, obj):
         """Register obj with the current transaction manager.
 
@@ -1016,6 +992,10 @@ class Connection(ExportImport, object):
         if obj is not None:
             self._registered_objects.append(obj)
 
+    def readCurrent(self, ob):
+        assert ob._p_jar is self
+        assert ob._p_oid is not None and ob._p_serial is not None
+        self._readCurrent[ob._p_oid] = ob._p_serial
 
     # persistent.interfaces.IPersistentDatamanager
     ##########################################################################
@@ -1089,10 +1069,12 @@ class Connection(ExportImport, object):
         if getattr(self, '_reader', None) is not None:
             self._reader._cache = cache
 
-    def _releaseStorage(self):
-        """Tell the storage to release resources it's using"""
-        if self._mvcc_storage:
-            self._storage.release()
+    def _release_resources(self):
+        for c in self.connections.itervalues():
+            if c._mvcc_storage:
+                c._storage.release()
+            c._storage = c._normal_storage = None
+            c._cache = PickleCache(self, 0, 0)
 
     ##########################################################################
     # Python protocol
@@ -1159,7 +1141,10 @@ class Connection(ExportImport, object):
         self._abort()
         self._registered_objects = []
         src = self._storage
-        self._invalidate_creating(src.creating)
+
+        # Invalidate objects created *after* the savepoint.
+        self._invalidate_creating((oid for oid in src.creating
+                                   if oid not in state[2]))
         index = src.index
         src.reset(*state)
         self._cache.invalidate(index)
@@ -1170,7 +1155,6 @@ class Connection(ExportImport, object):
         src = self._savepoint_storage
         self._storage = self._normal_storage
         self._savepoint_storage = None
-
         try:
             self._log.debug("Committing savepoints of size %s", src.getSize())
             oids = src.index.keys()
@@ -1200,7 +1184,7 @@ class Connection(ExportImport, object):
                     s = self._storage.store(oid, serial, data,
                                             '', transaction)
 
-                self._handle_serial(s, oid, change=False)
+                self._handle_serial(oid, s, change=False)
         finally:
             src.close()
 
@@ -1402,3 +1386,23 @@ class RootConvenience(object):
         if len(names) > 60:
             names = names[:57].rsplit(' ', 1)[0] + ' ...'
         return "<root: %s>" % names
+
+large_object_message = """The %s
+object you're saving is large. (%s bytes.)
+
+Perhaps you're storing media which should be stored in blobs.
+
+Perhaps you're using a non-scalable data structure, such as a
+PersistentMapping or PersistentList.
+
+Perhaps you're storing data in objects that aren't persistent at
+all. In cases like that, the data is stored in the record of the
+containing persistent object.
+
+In any case, storing records this big is probably a bad idea.
+
+If you insist and want to get rid of this warning, use the
+large_record_size option of the ZODB.DB constructor (or the
+large-record-size option in a configuration file) to specify a larger
+size.
+"""
