@@ -110,6 +110,7 @@ class ZODBDatabase(BaseConfig):
 
         _option('pool_timeout')
         _option('allow_implicit_cross_references', 'xrefs')
+        _option('large_record_size')
 
         try:
             return ZODB.DB(
@@ -157,8 +158,15 @@ class FileStorage(BaseConfig):
         config = self.config
         options = {}
         if getattr(config, 'packer', None):
-            m, name = config.packer.rsplit('.', 1)
-            options['packer'] = getattr(__import__(m, {}, {}, ['*']), name)
+            packer = config.packer
+            if ':' in packer:
+                m, expr = packer.split(':', 1)
+                m = __import__(m, {}, {}, ['*'])
+                options['packer'] = eval(expr, m.__dict__)
+            else:
+                m, name = config.packer.rsplit('.', 1)
+                m = __import__(m, {}, {}, ['*'])
+                options['packer'] = getattr(m, name)
 
         for name in ('blob_dir', 'create', 'read_only', 'quota', 'pack_gc',
                      'pack_keep_old'):
@@ -188,6 +196,8 @@ class ZEOClient(BaseConfig):
             options['blob_cache_size'] = self.config.blob_cache_size
         if self.config.blob_cache_size_check is not None:
             options['blob_cache_size_check'] = self.config.blob_cache_size_check
+        if self.config.client_label is not None:
+            options['client_label'] = self.config.client_label
 
         return ClientStorage(
             L,
