@@ -19,7 +19,12 @@ from ZODB.blob import Blob
 from ZODB.DB import DB
 from ZODB.FileStorage import FileStorage
 from ZODB.tests.testConfig import ConfigTestBase
-from zope.testing import doctest
+
+import os
+if os.environ.get('USE_ZOPE_TESTING_DOCTEST'):
+    from zope.testing import doctest
+else:
+    import doctest
 
 import os
 import random
@@ -145,7 +150,7 @@ class BlobUndoTests(BlobTestBase):
         # the blob footprint object should exist no longer
         self.assertRaises(KeyError, root.__getitem__, 'blob')
         database.close()
-        
+
     def testUndo(self):
         database = DB(self._storage)
         connection = database.open()
@@ -178,7 +183,7 @@ class BlobUndoTests(BlobTestBase):
         blob.consumeFile('consume1')
         root['blob'] = blob
         transaction.commit()
-        
+
         transaction.begin()
         blob = root['blob']
         open('consume2', 'w').write('this is state 2')
@@ -279,7 +284,7 @@ class RecoveryBlobStorage(BlobTestBase,
         transaction.commit()
         self._dst.copyTransactionsFrom(self._storage)
         self.compare(self._storage, self._dst)
-    
+
 
 def gc_blob_removes_uncommitted_data():
     """
@@ -335,7 +340,7 @@ Works with savepoints too:
 
     >>> transaction.commit() # doctest: +ELLIPSIS
     Copied blob file ...
-    
+
     >>> root['blob2'].open().read()
     'test2'
 
@@ -343,6 +348,7 @@ Works with savepoints too:
     >>> logger.propagate = True
     >>> logger.setLevel(0)
     >>> logger.removeHandler(handler)
+    >>> handler.close()
 
     >>> database.close()
     """
@@ -589,6 +595,7 @@ def savepoint_commits_without_invalidations_out_of_order():
     Traceback (most recent call last):
         ...
     ConflictError: database conflict error...
+    >>> tm1.abort()
     >>> db.close()
     """
 
@@ -619,7 +626,22 @@ def savepoint_cleanup():
 
     >>> db.close()
     """
+def lp440234_Setting__p_changed_of_a_Blob_w_no_uncomitted_changes_is_noop():
+    r"""
+    >>> conn = ZODB.connection('data.fs', blob_dir='blobs')
+    >>> blob = ZODB.blob.Blob('blah')
+    >>> conn.add(blob)
+    >>> transaction.commit()
+    >>> old_serial = blob._p_serial
+    >>> blob._p_changed = True
+    >>> transaction.commit()
+    >>> blob.open().read()
+    'blah'
+    >>> old_serial == blob._p_serial
+    True
 
+    >>> conn.close()
+    """
 
 def setUp(test):
     ZODB.tests.util.setUp(test)
@@ -654,7 +676,7 @@ def storage_reusable_suite(prefix, factory,
             return factory(name, blob_dir)
 
         test.globs['create_storage'] = create_storage
-    
+
     suite = unittest.TestSuite()
     suite.addTest(doctest.DocFileSuite(
         "blob_connection.txt", "blob_importexport.txt",
